#  Authors     [Robert Meolic (robert@meolic.com)]
#  Revision    [$Revision: 188 $]
#  Date        [$Date: 2019-02-07 15:11:19 +0100 (čet, 07 feb 2019) $]
#
#  Copyright   [This file is part of EST (Efficient Symbolic Tools).
#               Copyright (C) 2003, 2019
#               UM-FERI, Koroska cesta 46, SI-2000 Maribor, Slovenia
#
#               EST is free software; you can redistribute it and/or modify
#               it under the terms of the GNU General Public License as
#               published by the Free Software Foundation; either version 2
#               of the License, or (at your option) any later version.
#
#               EST is distributed in the hope that it will be useful,
#               but WITHOUT ANY WARRANTY; without even the implied warranty of
#               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#               GNU General Public License for more details.
#
#               You should have received a copy of the GNU General Public
#               License along with this program; if not, write to the Free
#               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
#               Boston, MA 02110-1301 USA.]

lappend auto_path [file dirname [info script]]

# HELP ON USAGE
# This file should be used from tclsh (use command 'source')
# This is initialization of EST package.

# ####################################################################
# define functions for output
# ####################################################################

global GUIOUTPUTLENGTH
set GUIOUTPUTLENGTH -1
global GUIOUTPUTLENGTHOK
set GUIOUTPUTLENGTHOK 0

# just a placeholder, gui_output should be redefined in C
proc gui_output line {
    puts -nonewline "CONSOLE>$line"
}

# out redirects output to C function (which uses printf)
proc out line {
    gui_output $line
}

# outline redirects output to C function (which uses printf)
proc outline line {
    global GUIOUTPUTLENGTH
    if {$GUIOUTPUTLENGTH >= 0} {
      incr GUIOUTPUTLENGTH 1
    }
    if {$line == ""} {
      gui_output ""
    } else {
      gui_output $line
      gui_output ""
    }
}

proc out_direct line { puts -nonewline $line }
proc outline_direct line { puts $line }

# ####################################################################
# useful functions from MyInterface package
# ####################################################################

variable timerStack {}
proc timerEnter {cmd op} {
    variable timerStack
    lappend timerStack [clock milliseconds]
}

#verbose
proc timerLeave {cmd code result op} {
    variable timerStack
    set now [clock milliseconds]
    set then [lindex $timerStack end]
    set timerStack [lrange $timerStack 0 end-1]
    if {[llength $timerStack] < 1 &&
        ![string equal -length 3 $cmd "out"] &&
        ![string equal -length 3 $cmd "set"] &&
        ![string equal -length 5 $cmd "global"] &&
        ![string equal -length 4 $cmd "expr"] &&
        ![string equal -length 4 $cmd "proc"]
    } {
      outline "\[CPU\]$cmd\[CPU: [expr {$now-$then}]ms\]"
    }
}

#very verbose!
proc timerLeaveX {cmd code result op} {
    variable timerStack
    set now [clock milliseconds]
    set then [lindex $timerStack end]
    set timerStack [lrange $timerStack 0 end-1]
    if {![string equal -length 3 $cmd "out"] &&
        ![string equal -length 5 $cmd "global"] &&
        ![string equal -length 4 $cmd "bind"] &&
        ![string equal -length 10 $cmd ".menuFrame"] &&
        ![string equal -length 7 $cmd ".output"] &&
        ![string equal -length 6 $cmd ".input"] &&
        ![string equal -length 5 $cmd "::tcl"] &&
        ![string equal -length 6 $cmd "update"] &&
        ![string equal -length 3 $cmd "eof"] &&
        ![string equal -length 4 $cmd "gets"] &&
        ![string equal -length 5 $cmd "catch"] &&
        ![string equal -length 8 $cmd "fblocked"] &&
        ![string equal -length 6 $cmd "myread"] &&
        ![string equal -length 13 $cmd "if \{\[fblocked"] &&
        ![string equal -length 8 $cmd "if \{\[eof"] &&
        ![string equal -length 4 $cmd "proc"]
    } {
      outline "\[CPU\]$cmd\[CPU: [expr {$now-$then}]ms\]"
    }
}

proc xsource fileName {
    global tcl_version

    if {$tcl_version < "8.5"} {
      outline "Error: xsource requires TCL version 8.5 (this is $tcl_version)"
      return
    }

    #This is from StackOverflow (thanks to Donal Fellows)
    trace add execution source enterstep timerEnter
    trace add execution source leavestep timerLeave

    source "$fileName"
    #use this for argument rewriting
    #source [set argv [lassign $argv argv0];set argv0]

    trace remove execution source enterstep timerEnter
    trace remove execution source leavestep timerLeave

    return
}

proc xxsource fileName {
    global tcl_version

    if {$tcl_version < "8.5"} {
      outline "Error: xxsource requires TCL version 8.5 (this is $tcl_version)"
      return
    }

    #This is from StackOverflow (thanks to Donal Fellows)
    trace add execution source enterstep timerEnter
    trace add execution source leavestep timerLeaveX

    source "$fileName"
    #use this for argument rewriting
    #source [set argv [lassign $argv argv0];set argv0]

    trace remove execution source enterstep timerEnter
    trace remove execution source leavestep timerLeaveX

    return
}

# ####################################################################
# useful function from tcl/tk part of bdd package
# ####################################################################

proc bdd_stat { } {
  outline "Statistic about BDD package"
  outline "NODE TABLE:"
  out "  bdd_table_number: "
  outline [bdd_table_number]
  out "  bdd_table_max: "
  outline [bdd_table_max]
  out "  bdd_list_used: "
  outline [bdd_list_used]
  out "  bdd_list_max_length: "
  outline [bdd_list_max_length]
  out "  bdd_list_avg_length: "
  outline [bdd_list_avg_length]
  outline "ITE CACHE:"
  out "  bdd_cache_ITE_search: "
  outline [bdd_cache_ITE_search]
  out "  bdd_cache_ITE_find: "
  outline [bdd_cache_ITE_find]
  out "  bdd_cache_ITE_overwrite: "
  outline [bdd_cache_ITE_overwrite]
  outline "RELOP CACHE:"
  out "  bdd_cache_RelOp_num: "
  outline [bdd_cache_RelOp_num]
  out "  bdd_cache_RelOp_search: "
  outline [bdd_cache_RelOp_search]
  out "  bdd_cache_RelOp_find: "
  outline [bdd_cache_RelOp_find]
  out "  bdd_cache_RelOp_overwrite: "
  outline [bdd_cache_RelOp_overwrite]
  outline "RELOPSIMPLE CACHE:"
  out "  bdd_cache_RelOpSimple_num: "
  outline [bdd_cache_RelOpSimple_num]
  out "  bdd_cache_RelOpSimple_search: "
  outline [bdd_cache_RelOpSimple_search]
  out "  bdd_cache_RelOpSimple_find: "
  outline [bdd_cache_RelOpSimple_find]
  out "  bdd_cache_RelOpSimple_overwrite: "
  outline [bdd_cache_RelOpSimple_overwrite]
  outline "RELOPCOMPLEX CACHE:"
  out "  bdd_cache_RelOpComplex_num: "
  outline [bdd_cache_RelOpComplex_num]
  out "  bdd_cache_RelOpComplex_search: "
  outline [bdd_cache_RelOpComplex_search]
  out "  bdd_cache_RelOpComplex_find: "
  outline [bdd_cache_RelOpComplex_find]
  out "  bdd_cache_RelOpComplex_overwrite: "
  outline [bdd_cache_RelOpComplex_overwrite]
  outline "STABRS CACHE:"
  out "  bdd_cache_StabRS_search: "
  outline [bdd_cache_StabRS_search]
  out "  bdd_cache_StabRS_find: "
  outline [bdd_cache_StabRS_find]
  out "  bdd_cache_StabRS_overwrite: "
  outline [bdd_cache_StabRS_overwrite]
  outline "MEMORY MANAGEMENT:"
  out "  bdd_block_number: "
  outline [bdd_block_number]
  out "  bdd_table_generated: "
  outline [bdd_table_generated]
  out "  bdd_garbage_number: "
  outline [bdd_garbage_number]
  out "  bdd_swap_number: "
  outline [bdd_swap_number]
  out "  bdd_sifting_number: "
  outline [bdd_sifting_number]
  outline ""
}

# ####################################################################
# Initialization
# ####################################################################

# in non-interactive mode we need an input file
if { $tcl_interactive == 0 } {
  if { $argc == 0 } {
    outline_direct "No input file passed in non-interactive mode!"
    return
  }
}

# ####################################################################
# GUI package
# ####################################################################

# This will start server.

# This will call gui_initPkg <countBytes> <host> <port>
# Please, use one of the following commands to start estsh2ed.tcl:
# % source estsh2ed.tcl #using server from estsh2ed, using port 9961
# % set ::GUIPORT 9961; source estsh2ed.tcl #using server from estsh2ed
# % set ::GUIHOST localhost; source estsh2ed.tcl #using meolic's server.tcl
# % set ::GUIHOST <host>; set ::GUIPORT 9961; source estsh2ed.tcl

# Counting bytes mechanism is only possible if server and client
# both support it, i.e. it works via a shared variable.
# You should not try to enable it if server does not support it!
# If using estsh2ed (internal server) and gui package (client)
# this mechanism is possible via GUIOUTPUTLENGTH and gui_outputLength.
# The mechanism is enabled by argument when calling function gui_initPkg.
# However, currently there are no advantages of using this mechanism.
# WARNING: Shared variable GUIOUTPUTLENGTH is also used to denote remote server!

if { [catch {
  package require est-server
  package require est-gui
  # if $::GUIPORT is not set then port 9961 will be used
  # if $::GUIHOST is not set then local server will be started and used
  if { [info exists ::GUIPORT] == 0 } {
    set ::GUIPORT 9961
  }
  if { [info exists ::GUIHOST] == 0 } {
    # mode 1 denotes local server without counting bytes mechanism
    # mode 2 denotes local server with counting bytes mechanism
    set mode 1
    set ::GUIHOST localhost
    socket -server [list est_accept [clock seconds]] $::GUIPORT
  } else {
    # mode 0 denotes remote server
    set mode 0 
    puts "USING REMOTE SERVER!"
  }
  gui_initPkg $mode $::GUIHOST $::GUIPORT
}] } {
  # in local mode output is direct to terminal and not via server
  # it is expected that EST is linked without gui module
  # and thus it is able to work without server
  rename out ""
  rename outline ""
  rename out_direct out
  rename outline_direct outline
  puts "THIS IS LOCAL MODE!"
}

# ####################################################################
# Say hello
# ####################################################################

outline "Efficient Symbolic Tools, 2nd Edition, Copyright (C) 2003-2019 UM-FERI"
outline "This is free software, and comes with ABSOLUTELY NO WARRANTY."
outline "You are welcome to redistribute EST; type \"license\" for details."
outline ""
outline "Running on $tcl_platform(machine) ($tcl_platform(os), $tcl_platform(osVersion)) with tcl $tcl_patchLevel."
outline ""

# ####################################################################
# BDD package
# ####################################################################

package require estsh-bdd
bdd_initPkg

# ####################################################################
# Process algebra package
# ####################################################################

package require estsh2ed-pa
pa_initPkg

# ####################################################################
# Versis package
# ####################################################################

package require estsh2ed-versis
versis_initPkg

# ####################################################################
# Model checking package
# ####################################################################

package require estsh2ed-mc
mc_initPkg

# ####################################################################
# Strucval package
# ####################################################################

package require estsh2ed-strucval
strucval_initPkg

# ####################################################################
# CCS package
# ####################################################################

package require estsh2ed-ccs
ccs_initPkg

# ####################################################################
# Verilog package
# ####################################################################

package require estsh2ed-verilog
verilog_initPkg

# ####################################################################
# OK. Let's go!
# ####################################################################

outline "Ready."
outline ""

if { $tcl_interactive == 0 } {
  if { $argc == 0 } {
    outline "ERROR: No input file in non-interactive mode!"
  } else {
    set i 1
    foreach arg $argv {
      if {$i > 0} {
        if { [file exists $arg] } {
          set dirname [file dirname $arg]
          set filename [file tail $arg]
          cd $dirname
          source $filename
        } else {
          outline "Warning: problem with file $arg"
        }
      }
      incr i
    }
  }
}

# THIS IS USED FOR DEBUGGING
# vwait forever

# ####################################################################
# License = GNU GPL (version 2)
# ####################################################################

proc license {  } {

  outline "Efficient Symbolic Tools, 2nd Edition, Copyright (C) 2003-2019 UM-FERI"

	outline ""
	outline "EST is free software; you can redistribute it and/or modify"
	outline "it  under  the  terms of  the GNU General Public License as"
	outline "published by the Free Software Foundation; either version 2"
	outline "of the License, or (at your option) any later version."

	outline ""
	outline "EST is distributed in the hope that it will be useful,  but"
	outline "WITHOUT  ANY  WARRANTY;  without even the implied  warranty"
	outline "of MERCHANTABILITY or FITNESS  FOR  A  PARTICULAR  PURPOSE."
	outline "See the GNU General Public License for more details."

	outline ""
	outline "You  should  have received a copy of the GNU General Public"
	outline "License along with this program;  if not, write to the Free"
	outline "Software Foundation, Inc., 51 Franklin Street,  Fifth Floor"
	outline "Boston, MA 02110-1301 USA."

	outline ""
}
